﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Interop;
using System.Windows.Media;
using Excel = Microsoft.Office.Interop.Excel;

namespace Demo
{

    public static class Extensions
    {
        public static string Description(this Enum value)
        {
            var @field = value?.GetType().GetField(value?.ToString());
            object[] attributes = @field?.GetCustomAttributes(typeof(DescriptionAttribute), false);
            return attributes.Length > 0 ? ((DescriptionAttribute)attributes?[0]).Description : (value?.ToString());
        }

        internal static string ToDescription(this Win32.DPI_AWARENESS_CONTEXT value)
        {
            switch (true)
            {
                case object _ when value.IsEqual(Win32.DPI_AWARENESS_CONTEXT.PER_MONITOR_AWARE_V2):
                    {
                        return Win32.DPI_AWARENESS_CONTEXT.PER_MONITOR_AWARE_V2.Description();
                    }
                case object _ when value.IsEqual(Win32.DPI_AWARENESS_CONTEXT.PER_MONITOR_AWARE):
                    {
                        return Win32.DPI_AWARENESS_CONTEXT.PER_MONITOR_AWARE.Description();
                    }
                case object _ when value.IsEqual(Win32.DPI_AWARENESS_CONTEXT.SYSTEM_AWARE):
                    {
                        return Win32.DPI_AWARENESS_CONTEXT.SYSTEM_AWARE.Description();
                    }
                case object _ when value.IsEqual(Win32.DPI_AWARENESS_CONTEXT.UNAWARE_GDISCALED):
                    {
                        return Win32.DPI_AWARENESS_CONTEXT.UNAWARE_GDISCALED.Description();
                    }
                case object _ when value.IsEqual(Win32.DPI_AWARENESS_CONTEXT.UNAWARE):
                    {
                        return Win32.DPI_AWARENESS_CONTEXT.UNAWARE.Description();
                    }
                case object _ when value.IsEqual(Win32.DPI_AWARENESS_CONTEXT.INVALID):
                    {
                        return Win32.DPI_AWARENESS_CONTEXT.INVALID.Description();
                    }

                default:
                    {
                        return "Undefined awareness!!!";
                    }
            }
        }

        internal static bool IsEqual(this Win32.DPI_AWARENESS_CONTEXT dacA, Win32.DPI_AWARENESS_CONTEXT dacB)
        {
            return Win32.AreDpiAwarenessContextsEqual(dacA, dacB);
        }
        public static T GetAssemblyAttribute<T>(this Assembly assy) where T : Attribute
        {
            object[] attr = assy.GetCustomAttributes(typeof(T), false);
            if (attr is null || attr.Length == 0)
                return null;
            return attr.OfType<T>().SingleOrDefault();
        }

        /// <summary>
    /// Get the Handle of a window.
    /// </summary>
        public static IntPtr Handle(this Window wnd)
        {
            return (wnd?.GetSource()?.Handle) ?? IntPtr.Zero;
        }

        /// <summary>
    /// Get the HwndSource of a window.
    /// </summary>
        public static HwndSource GetSource(this Window wnd)
        {
            if (wnd is null)
                return null;
            return (HwndSource)PresentationSource.FromVisual(wnd);
        }

        /// <summary>
    /// Get the HwndTarget of a window.
    /// </summary>
        public static HwndTarget GetTarget(this Window wnd)
        {
            return wnd?.GetSource()?.CompositionTarget;
        }

        /// <summary>
    /// Get the first (and sole?) Child of a window.
    /// </summary>
        public static FrameworkElement GetChild(this Window wnd)
        {
            if (wnd is null || VisualTreeHelper.GetChildrenCount(wnd) == 0)
                return null;
            return (FrameworkElement)VisualTreeHelper.GetChild(wnd, 0);
        }


        /// <summary>
    /// Helper to see is Excel Application is ready for manipulation,
    /// or blocked while user is 'in edit mode' or has a modal dialog open.
    /// </summary>
        public static bool IsReady(this Excel.Application app)
        {
            // doesn't work when non-modal names dialog is open
            try
            {
                return app is not null && app.Ready && app.BarReady();
            }
            catch 
            {
                // Call was rejected by callee
                Debug.Print("IsReady rejected:ignore!");
                return true;
            }
        }

        /// <summary>
    /// Attempts to get control of Excel by escaping edit mode
    /// </summary>
        public static bool DoReady(this Excel.Application app)
        {
            if (app is null)
            {
                return false;
            }
            else if (app.IsReady())
            {
                return true;
            }
            else
            {
                switch (Win32.GetFocusClassName() ?? "")
                {
                    case "NUL":
                    case "WND":
                    case "FRM":
                    case "XLMAIN":
                    case "XLDESK":
                    case "EXCEL7":
                        {
                            return true;
                        }
                    case "EXCEL<":
                    case "EXCEL=":
                    case "EXCEL6":
                        {
                            Win32.SendEsc();     // Escape Excel editing
                            Win32.SendClick();   // Click to refocus
                            return true;
                        }

                    default:
                        {
                            return false;
                        }
                }
            }
        }

        private static bool BarReady(this Excel.Application app)
        {
            // Mostly reliable, except for some non-modal dialogs. (FunctionWiz,NameMgr)
            try
            {
                if (app is null)
                    return false;
                return app.CommandBars.GetEnabledMso("FileNewDefault");
            }
            catch 
            {
                return false;
            }
        }
    }
}