﻿using System;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows;
using System.Windows.Threading;
using static Demo.Win32;
using Excel = Microsoft.Office.Interop.Excel;

namespace Demo
{

    public static class Library
    {
        // Demo of techniques/workarounds
        // DpiAwareness and Scaling in 'hosted' Wpf content
        // Modeless keyboard input

        public static bool IsHosted { get; set; }
        public static bool AllowFocusHelper { get; set; } = true;
        public static bool AllowDpiHelper { get; set; } = true;


        public static Application WpfApp { get; private set; }
        public static Excel.Application XlsApp { get; private set; }
        public static FocusHelper Focus { get; private set; }
        public static ViewModel ViewModel { get; private set; }


        public static void DisposeLib()
        {
            ShutExcel();
            WpfApp?.Dispatcher.InvokeShutdown();
        }

        public static void Show(object hostApp)
        {
            IsHosted = hostApp is not null;

            EnsureAwareness();
            EnsureApp();
            EnsureData();
            EnsureResources();
            EnsureMain();

            if (IsHosted)
            {
                InitExcel((Excel.Application)hostApp);
            }
            WpfApp.MainWindow.Show();
        }


        public static void EnsureAwareness()
        {

            if (IsProcessV2())
            {
                Debug.Print("Awareness:OK");
            }
            else if (!IsProcessV2())
            {
                Debug.Print($"Awareness:OLD {GetDpiData()}");
                // When WPF is Hosted (in my case in VSTO/Excel), access to SetProcessDpiAwareness is denied.

                // However:
                // HwndSource.cs, HwndTarget.cs and DpUtil.cs appear hardwired to SystemDPI on the first call to ProcessDpiAwareness.
                // and will not raise DPIchange events.

                // We can compensate for the most part by setting a LayoutTransform on the RootVisual and it's child element(s). 
                // However PopupRoot scaling will be based on SystemDPI and positioning is haphazerd.
                /* TODO ERROR: Skipped DefineDirectiveTrivia
                #Const curious = 0
                *//* TODO ERROR: Skipped IfDirectiveTrivia
                #If curious Then
                *//* TODO ERROR: Skipped DisabledTextTrivia
                            Dim pda(2) As PROCESS_DPI_AWARENESS
                            Dim pca(2) As DPI_AWARENESS_CONTEXT
                            Dim res(2) As Integer
                            res(0) = GetProcessDpiAwareness(IntPtr.Zero, pda(0))
                            pca(0) = GetDpiAwarenessContextForProcess()
                            res(0) = SetProcessDpiAwareness(PROCESS_DPI_AWARENESS.PROCESS_PER_MONITOR_DPI_AWARE)
                            res(1) = GetProcessDpiAwareness(IntPtr.Zero, pda(1))
                            pca(1) = GetDpiAwarenessContextForProcess()
                *//* TODO ERROR: Skipped EndIfDirectiveTrivia
                #End If
                */            // Nota bene: After setting the ThreadDAC the ProcessDAC=ThreadDAC
                SetThreadDpiAwarenessContext(DPI_AWARENESS_CONTEXT.PER_MONITOR_AWARE_V2);
                SetThreadDpiHostingBehavior(DPI_HOSTING_BEHAVIOR.MIXED);
                /* TODO ERROR: Skipped IfDirectiveTrivia
                #If curious Then
                *//* TODO ERROR: Skipped DisabledTextTrivia
                            res(2) = GetProcessDpiAwareness(IntPtr.Zero, pda(2))
                            pca(2) = GetDpiAwarenessContextForProcess()
                *//* TODO ERROR: Skipped EndIfDirectiveTrivia
                #End If
                */
                Debug.Print($"Awareness:NEW {GetDpiData()}");
                if (!IsProcessV2())
                    throw new Exception("Redmond!.. can't set awareness!");
            }
        }

        public static void EnsureApp()
        {
            if (Application.Current is null)
                LoadApp();
        }

        public static void EnsureResources()
        {
            // I'm a library, so no application framework App.xaml (c#) or Application.xaml (VB)
            // thus no Application Resources!

            // Create and initialize (load) it here.
            EnsureApp();
            LoadDefaultResources();
        }
        public static void EnsureData()
        {
            if (ViewModel is null)
                ViewModel = new ViewModel();
        }
        public static void EnsureMain()
        {
            if (WpfApp is null)
                Debugger.Break();
            if (WpfApp.MainWindow is null)
            {
                Application.Current.MainWindow = new PlainWindow();
            }
        }


        public static void LoadApp()
        {
            if (!IsProcessV2())
                Debugger.Break();
            if (IsHosted)
            {
                WpfApp = new Application() { ShutdownMode = ShutdownMode.OnExplicitShutdown };
            }
            else
            {
                WpfApp = new Application() { ShutdownMode = ShutdownMode.OnMainWindowClose };
            }

            if (WpfApp is not null)
            {
                WpfApp.DispatcherUnhandledException += WpfApp_DispatcherUnhandledException;
                WpfApp.Startup += WpfApp_Startup;
            }
        }

        public static void LoadDefaultResources()
        {
            {
                var withBlock = Application.Current.Resources.MergedDictionaries;
                if (withBlock.Count > 0)
                    return;
                LoadResources("StylesDefault", "DpiDemo.Lib");
            }
        }

        // DpiDemo.Lib / DpiDemo.LibAP
        // StylesCustom / StylesBarCustom
        public static void LoadResources(string name, string assy)
        {
            {
                var withBlock = Application.Current.Resources.MergedDictionaries;
                if (withBlock.SingleOrDefault(e => (e.Source?.OriginalString.Contains(name)) ?? false) is null)
                {
                    withBlock.Add(new ResourceDictionary() { Source = new Uri($"pack://application:,,,/{assy};component/Resources/{name}.xaml") });
                }
            }
        }

        public static void UnloadResources(string name)
        {
            {
                var withBlock = Application.Current.Resources.MergedDictionaries;
                var find = withBlock.SingleOrDefault(e => (e.Source?.OriginalString.Contains(name)) ?? false);
                if (find is not null)
                    withBlock.Remove(find);
            }
        }


        public static void InitExcel(Excel.Application xlApp)
        {
            if (xlApp is null)
                return;
            if (Application.Current is null)
                throw new Exception("Wpf App must be initialized before Excel.");
            if (WpfApp is null)
                WpfApp = Application.Current;


            XlsApp = xlApp;
            Focus = new FocusHelper(WpfApp.MainWindow);
            Focus.Show();
        }

        public static void ShutExcel()
        {
            if (Focus is not null)
            {
                Focus.Dispose();
                Focus = null;
            }
            if (XlsApp is not null)
            {
                Marshal.FinalReleaseComObject(XlsApp);
                XlsApp = null;
            }

        }
        public static bool IsProcessV2()
        {
            return GetDpiAwarenessContextForProcess().IsEqual(DPI_AWARENESS_CONTEXT.PER_MONITOR_AWARE_V2);
        }
        public static (string Process, string Thread, string Hosting) GetDpiData()
        {
            return (GetDpiAwarenessContextForProcess().ToDescription(), GetThreadDpiAwarenessContext().ToDescription(), GetThreadDpiHostingBehavior().Description());
        }


        private static void WpfApp_Startup(object sender, StartupEventArgs e)
        {
            Debug.Print($"WpfApp:StartUp:{(e.Args.Length > 0 ? e.Args[0] : "no args")}");
            Debug.Print($"WpfApp:StartUp:IsHosted:{IsHosted} IsV2:{IsProcessV2()}");
        }
        private static void WpfApp_DispatcherUnhandledException(object sender, DispatcherUnhandledExceptionEventArgs e)
        {
            Debug.Print($"WpfApp:Unhandled:{e.Exception.Message}");
            e.Handled = true;
        }






    }
}